/*
  clp.pl

  @author Francois Fages
  @version 1.1.3 

  @email Francois.Fages@inria.fr
  @license LGPL-2

  Frontend to library(clpr) and library(clpfd) to allow shorthand functional notations in clpr and clpfd constraints on arrays and lists

*/

:- module(clp, [
		truth_value/2,

				% CLPFD CONSTRAINTS GENERALIZED TO HYBRID CLPFD CLPR CONSTRAINTS
		sum/3,
		scalar_product/4,

				% NEW HYBRID CLPFD CLPR CONSTRAINTS
		op_rel/5,
		rel/3,

				% NEW CLPR CONSTRAINTS
		float_sum/3,
		float_product/3,

				% NEW CLPFD CONSTRAINTS
		int_sum/3,
		int_product/3,
		int_minimum/3,
		int_minimum/2,
		int_maximum/3,
		int_maximum/2,
		
		lex_leq/2,
		lex_lt/2,

		
				% CLPFD CONSTRAINTS GENERALIZED TO ARRAYS
                in/2,
                ins/2,
		(#>)/2,
		(#<)/2,
		(#>=)/2,
		(#=<)/2,
		(#=)/2,
		(#\=)/2,
		(#\)/1,
		(#<==>)/2,
		(#==>)/2,
		(#<==)/2,
		(#\/)/2,
		(#\)/2,
		(#/\)/2,
		
		all_different/1,
		all_distinct/1,
		tuples_in/2,
		labeling/2,
		label/1,
		lex_chain/1,
		serialized/2,
		global_cardinality/2,
		global_cardinality/3,
		circuit/1,
		cumulative/1,
		cumulative/2,
		disjoint2/1,
		element/3,
		automaton/3,
		transpose/2,
		chain/2,

		%op(450, xfx, ..), % in clpfd: "should bind more tightly than \/ of priority 500"
		op(501, xfx, ..), % but here less tightly than arithmetic expressions now allowed in domains where * has priority 500
		op(502, yfx, \/), % consequently changed from 500 to 502 WITH ONE RISK FOR BITWISE OR expressions !!!

		op(760, yfx, #<==>),
		op(750, xfy, #==>),
		op(750, yfx, #<==),
		op(740, yfx, #\/),
		op(730, yfx, #\),
		op(720, yfx, #/\),
		op(710,  fy, #\),
		op(700, xfx, #>),
		op(700, xfx, #<),
		op(700, xfx, #>=),
		op(700, xfx, #=<),
		op(700, xfx, #=),
		op(700, xfx, #\=),
		op(700, xfx, in), 
		op(700, xfx, ins),


				% CLPR

		float_sum/3,
		float_product/3,
		
		'{=}'/2,
		'{>=}'/2,
		'{>}'/2,
		'{<}'/2,
		'{=<}'/2,
		'{=\\=}'/2,
		
		{}/1,
		entailed/1,
		maximize/1,
		minimize/1,
		inf/2,
		sup/2, 
		bb_inf/3,
		bb_inf/5,


				% ARRAYS
		op(100, yf, []) % for array cell functional notation Array[Indices]
		]).

/** <module> Frontend to library(clpr) and library(clpfd) to allow shorthand functional notations in clpr and clpfd constraints on arrays and lists

@author Francois Fages
@version 1.1.3
  
This library
  
  * defines a few extra clpfd constraints int_sum/3, int_product/3, int_minimum/2, int_maximum/2, lex_leq/2, lex_lt/2,
  and clpr constraints float_sum/3, float_product/3;
  
  * adds reification constraint truth_value/2 for both clpfd and clpr constraints

==
?- array(A, [3]), truth_value({A[1] < 3.14}, Boolean).
A = array(_A, _, _),
when((nonvar(_A);nonvar(B)), clp:clpr_reify(_A<3.14, _A>=3.14, [_A], Boolean)),
clpfd:(Boolean in 0..1).

?- array(A, [3]), truth_value({A[1] < 3.14}, Boolean), array_list(A, [1, 2.7, 42]).
A = array(1, 2.7, 42),
Boolean = 1.
==
  
  * allows user:shorthand/3 functional notations in domains and constraints,
  e.g. for subscripted variables with Array[Indices] functional notation:

==
?- array(A, [3]), for_all(I in 1..2, A[I] #< A[I+1]).
A = array(_A, _B, _C),
clpfd:(_A#=<_B+ -1),
clpfd:(_B#=<_C+ -1).
==

  * adds user:shorthand/3 clauses for allowing constraint predicates to appear as constraint functions in expressions, with last predicate argument as result,
  
  e.g. the magic series puzzle can be writen using shorthand truth_value functional notation for truth_value/2 constraint:
==
?- array(X, [N]), for_all(I in 1..N, X[I] #= int_sum(J in 1..N, truth_value(X[J] #= I-1))), label(X).
X = array(1, 2, 1, 0),
N = 4 ;
X = array(2, 0, 2, 0),
N = 4 ;
X = array(2, 1, 2, 0, 0),
N = 5 ;
X = array(3, 2, 1, 1, 0, 0, 0),
N = 7 ;
X = array(4, 2, 1, 0, 1, 0, 0, 0),
N = 8 ;
X = array(5, 2, 1, 0, 0, 1, 0, 0, 0),
N = 9 ;
X = array(6, 2, 1, 0, 0, 0, 1, 0, 0, 0),
N = 10 .
==

  * defines hybrid clpr-clpfd constraints on lists and arrays, including op_rel/5, sum/3, e.g.

==
?- A=array([1,2],[3,4]), op_rel(A, +, A, #=, C).
A = array([1, 2], [3, 4]),
C = array(array(2, 4), array(6, 8)).

?- A=[[1,2],[3,4]], op_rel(A, +, A, #=, C).
A = [[1, 2], [3, 4]],
C = [[2, 4], [6, 8]].

?- L=[X, Y, Z], L ins 1..3, for_all(V in L, V#>1).
L = [X, Y, Z],
clpfd:(X in 2..3),
clpfd:(Y in 2..3),
clpfd:(Z in 2..3).
==

  * adds option trace to labeling/2 predicate for building the search tree and visualizing it using search_tree_text/0, search_tree_term/1, search_tree_latex/1, search_tree_tikz/1 of library(tracesearch).

  * adds a patch to library(clpfd) unify_attributes_hook/2 to avoid the duplication of constraint propagators when unifying two clpfd variables.

  The duplication of library(clpfd) constraints was due to the use of append/3 in unify_attributes_hook/2 now replaced by union_propagators/3 

==  
?- use_module(library(clpfd)).
true.

?- X#>Z, Y#>Z, X#=Y.
X = Y,
Z#=<Y+ -1,
Z#=<Y+ -1.

?- L=[A, B], L ins 1..2, A #=< B, bagof(W, member(W, L), L2).
L = L2, L2 = [A, B],
A in 1..2,
B#>=A,
B#>=A,
B#>=A,
B in 1..2.
==

You can set nb_setval(fdvar_unification_warning, true) to get warnings when two fd variables are unified (default is false).

TODO: the constraint propagators should rather be simplified when 2 variable unify (as done in SMT solvers).
This is a winning strategy on global constraints compared to mere domain consistency checking;

e.g. simplification of sum scalar_product lex_chain serialized global_cardinality cumulative element automaton? transpose? chain?

e.g. in all_distinct/1 all_different/1 unify_attribute_hook should rather fail when propagators get duplicated variables

==
?- L=[X, Y, Z], L ins 1..3, all_distinct(L), X #= Y.
Warning: unifying fdvar _252 clpfd_attr(no,no,no,from_to(n(1),n(3)),fd_props([propagator(pexclude([_252,_308],[],_252),_396)],[],[propagator(pdistinct([_252,_252,_308]),_554)]))
 with fdvar domain from_to(n(1),n(3)) fd_props([propagator(pexclude([_308],[_252],_252),_466)],[],[propagator(pdistinct([_252,_252,_308]),_554)])
L = [Y, Y, Z],
X = Y,
clpfd:(Y in 1..3),
clpfd:all_distinct([Y, Y, Z]),
clpfd:(Z in 1..3).
==
*/


				% PATCH TO CLPFD LIBRARY FOR SHORTHAND FUNCTIONAL NOTATIONS, CONSTRAINTS ON ARRAYS AND EXTRA CONSTRAINTS


% Just for pldoc 
:- op(501, xfx, ..).
:- op(502, yfx, \/).
:- op(700, xfx, in).
:- op(700, xfx, ins).

:- reexport(library(arrays)).

:- reexport(library(tracesearch)).

:- nb_setval(labeling_variable_names_reversed, []).

:- nb_setval(fdvar_unification_warnings, false). % to set to true to get a warning when two clpfd variables unify


:- use_module(library(clpfd),
	      except([
		      (#>)/2,
		      (#<)/2,
		      (#>=)/2,
		      (#=<)/2,
		      (#=)/2,
		      (#\=)/2,
		      (#\)/1,
		      (#<==>)/2,
		      (#==>)/2,
		      (#<==)/2,
		      (#\/)/2,
		      (#\)/2,
		      (#/\)/2,
		      in/2,
		      ins/2,
		      all_different/1,
		      all_distinct/1,
		      sum/3,
		      scalar_product/4,
		      tuples_in/2,
		      labeling/2,
		      label/1,
		      lex_chain/1,
		      serialized/2,
		      global_cardinality/2,
		      global_cardinality/3,
		      circuit/1,
		      cumulative/1,
		      cumulative/2,
		      disjoint2/1,
		      element/3,
		      automaton/3,
		      transpose/2,
		      chain/2,
		      op(450, xfx, ..),
		      op(500, xfx, \/)
		     ]
		    )).

:- op(501, xfx, ..).
:- op(502, yfx, \/).

:- use_module(library(clpr),
	      except([
	      	      {}/1,
	      	      entailed/1,
	      	      maximize/1,
	      	      minimize/1,
	      	      inf/2,
	      	      sup/2, 
	      	      bb_inf/3,
	      	      %bb_inf/4,
	      	      bb_inf/5
	      	     ]
	      )).

    
				% NEW CLPFD CONSTRAINTS ADDED


%! int_sum(+VarDomains, +Expr, ?Sum)
% 
% Sum of int Expr defined in comprehension by "in" domain and "where" condition

int_sum(VarDomains, Expr, Sum):-
    aggregate_for(VarDomains, Expr, +, 0, #=, Sum).


%! int_product(+VarDomains, +Expr, ?Product)
% 
% Product of int Expr defined in comprehension by "in" domain and "where" condition

int_product(VarDomains, Expr, Product):-
    aggregate_for(VarDomains, Expr, *, 1, #=, Product).


%! int_minimum(+A, ?Min)
% 
% constrains variable Min to be the minimum value of non-empty array or list A of integers.

int_minimum(A, Min):- 
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, List) ; must_be(list, Ae), flatten(Ae, List)),
    (List=[X0 | Tail]
    ->
     aggregate_for([X in Tail], X, min, X0, #=, Min)
    ;
     throw(error(minimum_of_empty_set(A)))).


%! int_minimum(+VarDomains, +Expr, ?Minimum)
% 
% Minimum value of int Expr in non-empty set defined in comprehension by "in" domain and "where" conditions.

int_minimum(VarDomains, Expr, Min):-
    list_of(VarDomains, Expr, List),
    int_minimum(List, Min).



%! int_maximum(+VarDomains, +Expr, ?Maximum)
% 
% Maximum value of int Expr in non-empty set defined in comprehension by "in" domain and "where" conditions.

int_maximum(VarDomains, Expr, Max):-
    list_of(VarDomains, Expr, List),
    int_maximum(List, Max).


%! int_maximum(+A, ?Max)
% 
% constrains variable Max to be the maximum value in non-empty array or list A of integers.

int_maximum(A, Max):- 
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, List) ; must_be(list, Ae), flatten(Ae, List)),
    (List=[X0 | Tail]
    ->
     aggregate_for([X in Tail], X, max, X0, #=, Max)
    ;
     throw(error(maximum_of_empty_set(A)))).


%! lex_leq(+A, +B)
%
% lexicographic ordering between arrays or lists

lex_leq(A, B):-
    expand(A, Ae),
    expand(B, Be),
    (is_list(Ae) -> Alist=Ae ; array_list(Ae, Alist)),
    (is_list(Be) -> Blist=Be ; array_list(Be, Blist)),
    lex_leq_list(Alist, Blist).
    

lex_leq_list([], _).
lex_leq_list([X|L], [Y|M]) :-
    X#=<Y,
    freeze(X, freeze(Y, (X==Y -> lex_leq_list(L, M) ; true))).


%! lex_lt(+A, +B)
%
% strict lexicographic ordering between arrays or lists

lex_lt(A, B):-
    expand(A, Ae),
    expand(B, Be),    
    (is_list(Ae) -> Alist=Ae ; array_list(Ae, Alist)),
    (is_list(Be) -> Blist=Be ; array_list(Be, Blist)),
    lex_lt_list(Alist, Blist).
    

lex_lt_list([], [_ | _]).
lex_lt_list([X|L], [Y|M]) :-
    X#=<Y,
    freeze(X, freeze(Y, (X==Y -> lex_lt_list(L, M) ; true))).



				% CLPFD DOMAINS WITH EVALUABLE EXPRESSIONS AND SHORTHAND NOTATIONS

% TODO generalize to hybrid clpr domain constraints ?


%! ?Var in +Domain
% 
% Var is in Domain which can be a list or array of domains allowing for evaluable arithmetic expressions including array notations.

Var in Domain :-
    evaluate_domain(Domain, Dom),
    expand(Var, V),
    clpfd:(V in Dom).


%! +Vars ins +Domain
% 
% List or array Vars is in Domain which can be a list or array of domains allowing for evaluable arithmetic expressions including array notations.

Vars ins Domain :-
    expand(Vars,Varse),
    (array(Varse) -> array_list(Varse, List) ; List=Varse),
    evaluate_domain(Domain, Dom),
    clpfd:(List ins Dom).


evaluate_domain(Domain, Dom):-
    expand(Domain, Doma),
    eval_domain(Doma, Dom).


eval_domain(Domain, Dom):-
    must_be(nonvar, Domain),
    Domain=Min..Max
    ->
    (Min = inf -> Mi = Min ; Mi is Min), 
    (Max = sup -> Ma = Max ; Ma is Max), 
    Dom = Mi..Ma
    ;
    Domain = (Domain1 \/ Domain2)
    ->
    eval_domain(Domain1, Dom1),
    eval_domain(Domain2, Dom2),
    Dom = (Dom1 \/ Dom2)
    ;
    is_list(Domain)
    ->
    list_to_union(Domain, Dom)
    ;
    array(Domain)
    ->
    array_list(Domain, List),
    list_to_union(List, Dom)
    ;
    Dom is Domain.


list_to_union([Domain], Dom):-
    !,
    eval_domain(Domain, Dom).

list_to_union([Domain | Tail], Dom\/Union):-
    eval_domain(Domain, Dom),
    list_to_union(Tail, Union).



				% REIFICATION OF CLPR CLPFD CONSTRAINTS


%! truth_value(+Constraint, ?Boolean)
%
% Reification constraint for clpr and clpfd constraints equivalent to Boolean #<==> Constraint for a clpfd constraint.
% ==
% ?- array(A, [3]), truth_value({A[1] < 3.14}, B).
% A = array(_A, _, _),
% when((nonvar(_A);nonvar(B)), clp:clpr_reify(_A<3.14, _A>=3.14, B)).
% ==

truth_value(Constraint, Boolean):-
    expand(Constraint, Ce),
    expand(Boolean, Be),
    (compound(Ce), Ce={C}
    ->
     clpr_reify(C, Be)
    ;
     clpfd:(Be #<==> Ce)).


% clpr_reify(+Constraint, ?Boolean)
%
% Reification of clpr Constraint (without curly brackets)  by clpfd 0-1 variable Boolean.
% Freezes the posting of Constraint (resp. its negation) until the Boolean gets instantiated to 1 (resp. 0).
% In the other direction, freezes clpr entailment check for Boolean instanciation upon clpr variable instanciation only.

clpr_reify(Constraint, Boolean):-
    clpr_not(Constraint, NotConstraint),
    clpr_reify(Constraint, NotConstraint, Boolean).

clpr_reify(Constraint, NotConstraint, Bool):-
    (var(Bool)
    ->
     (entailed(Constraint) -> Bool=1 ;
      entailed(NotConstraint) -> Bool=0 ;
      term_variables(Constraint, Vars),
      nonvar_cond(Vars, Bool, Cond),
      when(Cond, clpr_reify(Constraint, NotConstraint, Bool)))
    ;
     (Bool=1 -> {Constraint} ; {NotConstraint})).

nonvar_cond([], Bool, nonvar(Bool)).
nonvar_cond([X | Vars], Bool, Condition) :-
    Condition = (nonvar(X) ; Cond),
    nonvar_cond(Vars, Bool, Cond).


% TODO: find strategy for periodic clpr entailment checking with priority queue ?

% clpr_semi_reify(+Constraint, ?Boolean)
%
% Semi reification of clpr Constraint (given without curly brakets) with clpfd 0-1 variable Boolean.
% Freezes the posting of Constraint (resp. its negation) until the Boolean gets instantiated to 1 (resp. 0).
% In the other direction, the entailment of Constraint is checked once (without freeze) to instantiate the Boolean.

clpr_semi_reify(C, B):-
    clpr_not(C, NC),
    clpr_semi_reify(C, NC, B).


clpr_semi_reify(Constraint, NotConstraint, Bool):-
    (var(Bool)
    ->
     (entailed(Constraint) -> Bool=1 ;
      entailed(NotConstraint) -> Bool=0 ;
      freeze(Bool, clpr_semi_reify(Constraint, NotConstraint, Bool)))
    ;
     (Bool=1 -> {Constraint} ; {NotConstraint})).


% clpr_entailed(+Constraint, ?Boolean)
%
% Similar to truth_value/2 for clpr Constraint (given without curly brakets) but checking entailment only once, without any freeze.
% This predicate can be called periodically, in absence of good waking up conditions on the clpr constraint store.

clpr_entailed(Constraint, Bool):-
    expand(Constraint, C),
    expand(Bool, B),
    clpr_not(C, NC),
    clpr_entailed(C, NC, B).


clpr_entailed(Constraint, NotConstraint, Bool):-
    (var(Bool)
    ->
     (entailed(Constraint) -> Bool=1 ;
      entailed(NotConstraint) -> Bool=0 ;
      true)
    ;
     (Bool=1 -> {Constraint} ; {NotConstraint})).


% negation of clpr constraints

clpr_not(A < B, A >= B).
clpr_not(A > B, A =< B).
clpr_not(A =< B, A > B).
clpr_not(A >= B, A < B).
clpr_not(A =\= B, A = B).
clpr_not(A =:= B, A = B).
clpr_not(A = B, A =\= B).
clpr_not((C, D), (E; F)):-
    clpr_not(C, E),
    clpr_not(D, F).
clpr_not((C; D), (E, F)):-
    clpr_not(C, E),
    clpr_not(D, F).


				% SHORTHANDS FOR ALLOWING CONSTRAINTS IN EXPRESSIONS
				% WHEN THE LAST ARGUMENT DENOTES A RESULT



:- multifile user:shorthand/3.


%! user:shorthand(+Term, +Expanded, +Goal)
%
% Multifile shorthand/3 clauses defined here to allow 
% - all clpfd constraints without their last argument in expressions when that last argument can be interpreted as a result
% e.g. functional notations for constraints sum/3 and truth_value/2 in expressions like
% ==
% X[I] #= int_sum(J in 1..N, truth_value(X[J] #= I-1)).
% ==
% - functions from domain predicates  fd_dom/2, fd_min/2, fd_max/2 inside expressions.

user:shorthand(fd_dom(Var), Domain, clp:fd_dom(Var, Domain)) :- !.

user:shorthand(fd_inf(Var), Min, clp:fd_inf(Var, Min)) :- !.

user:shorthand(fd_sup(Var), Max, clp:fd_sup(Var, Max)) :- !.

user:shorthand(truth_value(Constraint), Boolean, clp:truth_value(Constraint, Boolean)) :- !.

user:shorthand(int_sum(VarDomains, Expr), Result, clp:int_sum(VarDomains, Expr, Result)) :- !.

user:shorthand(int_product(VarDomains, Expr), Product, clp:int_product(VarDomains, Expr, Product)) :- !.

user:shorthand(int_minimum(VarDomains, Expr), Min, clp:int_minimum(VarDomains, Expr, Min)) :- !.

user:shorthand(float_sum(VarDomains, Expr), Result, clp:float_sum(VarDomains, Expr, Result)) :- !.

user:shorthand(float_product(VarDomains, Expr), Product, clp:float_product(VarDomains, Expr, Product)) :- !.

user:shorthand(int_minimum(A), Min, clp:int_minimum(A, Min)) :- !.

user:shorthand(int_maximum(VarDomains, Expr), Max, clp:int_maximum(VarDomains, Expr, Max)) :- !.

user:shorthand(int_maximum(A), Max, clp:int_maximum(A, Max)) :- !.

user:shorthand(rel(Arg1, Rel), Arg2, clp:rel(Arg1, Rel, Arg2)) :- !.

user:shorthand(op_rel(Arg1, Op, Arg2, Rel), Arg3, clp:op_rel(Arg1, Op, Arg2, Rel, Arg3)) :- !.
    
user:shorthand(op_rel_list(AA, Op, AB, Rel), AC, clp:op_rel_list(AA, Op, AB, Rel, AC)) :- !.

user:shorthand(sum(A, Rel), B, clp:sum(A, Rel, B))  :- !.

user:shorthand(scalar_product(A, B, Rel), C, clp:scalar_product(A, B, Rel, C))  :- !.

user:shorthand(element(A, B), C, clp:element(A, B, C))  :- !.

user:shorthand(transpose(A), B, clp:transpose(A, B))  :- !.



				% FRONTEND TO CLPFD ARITHMETIC CONSTRAINTS

%! #\ X
%
% front-end to clpfd constraint to allow shorthand functional notations

#\ X :- expand(X, Xe), clpfd:(#\ Xe).                  

%! X #= Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #= Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #= Ye).                  


%! X #\= Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #\= Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #\= Ye).                  


%! X #<==> Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #<==> Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #<==> Ye).                  


%! X #> Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #> Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #> Ye).

%! X #< Y
%
%front-end to clpfd constraint to allow shorthand functional notations

X #< Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #< Ye).

%! X #>= Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #>= Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #>= Ye).

%! X #=< Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #=< Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #=< Ye).
           

%! X #==> Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #==> Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #==> Ye).

%! X #<== Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #<== Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #<== Ye).

%! X #\/ Y
%
%front-end to clpfd constraint to allow shorthand functional notations

X #\/ Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #\/ Ye).

%! X #\ Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #\ Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #\ Ye).

%! X #/\ Y
%
% front-end to clpfd constraint to allow shorthand functional notations

X #/\ Y :- expand(X, Xe), expand(Y, Ye), clpfd:(Xe #/\ Ye).



				% CLPFD GLOBAL CONSTRAINTS ON LISTS GENERALIZED TO ARRAYS

				% FIRST HYBRID CONSTRAINTS GENERALIZED TO BOTH CLPFD CLPR CONSTRAINTS


%! rel(?Arg1, +Relation, ?Arg2)
%
% applies binary Relation to Arg1 and Arg2, element-wise.
% The arguments can be arrays, lists, clpfd or clpr variables or constants interpreted as constant array or list of appropriate size.

rel(Arg1, Rel, Arg2):-
    op_rel(Arg1, +, 0, Rel, Arg2).


%! op_rel(?Arg1, +Operation, ?Arg2, +Relation, ?Arg3)
%
% applies binary Operation to Arg1 and Arg2, element-wise, and compares the results with Relation to Arg3.
% The arguments can be arrays, lists, clpfd or clpr variables or constants interpreted as constant array or list of appropriate size.
% 
% E.g. the Hadamart product of two integer matrices of same dimension can be defined by
% ``op_rel(A, *, B, #=, C)`` over integers,
% ``op_rel(A, *, B, '{=}', C) `` over real numbers.

op_rel(Arg1, Op, Arg2, Rel, Arg3):-
    expand(Arg1, Ae1),
    expand(Arg2, Ae2),
    expand(Arg3, Ae3),
    expand(Operation, Op),
    expand(Rel, Relation),
    (array(Ae1) -> array_lists(Ae1, T1) ; var(Ae1), \+ fd_var(Ae1) -> true; T1=Ae1),
    (array(Ae2) -> array_lists(Ae2, T2) ; var(Ae2), \+ fd_var(Ae2) -> true; T2=Ae2),
    (array(Ae3) -> array_lists(Ae3, T3) ; var(Ae3), \+ fd_var(Ae3) -> true; T3=Ae3),
    
    op_rel_list(T1, Operation, T2, Relation, T3),
    
    ((array(Ae1) ; array(Ae2) ; array(Ae3))
     ->
      (var(Ae1), \+ fd_var(Ae1) -> (is_list(T1) -> array_lists(Ae1, T1) ; Ae1=T1) ; true),
      (var(Ae2), \+ fd_var(Ae2) -> (is_list(T2) -> array_lists(Ae2, T2) ; Ae2=T2) ; true),
      (var(Ae3), \+ fd_var(Ae3) -> (is_list(T3) -> array_lists(Ae3, T3) ; Ae3=T3) ; true)
     ;
      (var(Ae1), \+ fd_var(Ae1) -> Ae1=T1 ; true),
      (var(Ae2), \+ fd_var(Ae2) -> Ae2=T2 ; true),
     (var(Ae3), \+ fd_var(Ae3) -> Ae3=T3 ; true)
    ).


    
op_rel_list(AA, Op, AB, Rel, AC):-
    ((is_list(AA) ; is_list(AB) ; is_list(AC))
    ->
     (\+ fd_var(AA), AA=[A | As] -> true ; A=AA, As=AA),
     (\+ fd_var(AB), AB=[B | Bs] -> true ; B=AB, Bs=AB),
     (\+ fd_var(AC), AC=[C | Cs] -> true ; C=AC, Cs=AC),
     
     op_rel(A, Op, B, Rel, C),
     
     ((tail(As) ; tail(Bs); tail(Cs))
     ->
      op_rel_list(As, Op, Bs, Rel, Cs)
     ;
      (var(As), As\==A -> As=[] ; true),
      (var(Bs), Bs\==B -> Bs=[] ; true),
      (var(Cs), Cs\==C -> Cs=[] ; true)
     )
    ;
     Arg =.. [Op, AA, AB],
     call(Rel, Arg, AC)
    ).

tail(T):-
    is_list(T),
    T \= [].


%! sum(+A, +Relation, ?Sum)
% 
% hybrid constraint to constrain Sum to be in clpfd or clpr Relation to the sum of elements in array or list A.

sum(A, Rel, B) :-
    expand(A, Ae),
    expand(B, Be),
    expand(Rel, Relation),
    (array(Ae) -> array_list(Ae, L) ; L=Ae),
    (array(Be) -> array_list(Be, M) ; M=Be),
    (
     clpr_rel(Relation)
    ->
     clp:clpr_sum(L, Relation, M)
    ;
     clpfd:sum(L, Relation, M)).


% clpr_sum(+L, +Rel, ?V)
% 
% constrains the sum of values in list L to be in clpr relation Rel with V.

clpr_sum(L, Rel, B) :-
    clpr_sum_list(L, Sum),
    Goal =.. [Rel, Sum, B],
    call(Goal).


clpr_sum_list([X | List], Sum):-
    (List=[]
    ->
     Sum=X
    ;
     {Sum = X+S},
     clpr_sum_list(List, S)
    ).


%! scalar_product(+A, +B, +Rel, ?C) 
%
% constraint on arrays or lists similar to library(clpfd) constraint scalar_product on list of integers or reals.

scalar_product(A, B, Rel, C) :-
    expand(A, Ae),
    expand(B, Be),
    expand(C, Ce),
    expand(Rel, Relation),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    (array(Be) -> array_list(Be, BL) ; BL=Be),
    (array(Ce) -> array_list(Ce, CL) ; CL=Ce),
    (clpr_rel(Relation)
    ->
     op_rel(AL, *, BL, '{=}', AB),
     sum(AB, Relation, CL)
    ;
     clpfd:scalar_product(AL, BL, Relation, CL)).


%! transpose(+A, ?B) 
%
% constraint on arrays or lists A and B similar to library(clpfd) constraint transpose on lists. 

transpose(A, B) :-
    expand(A, Ae),
    expand(B, Be),
    (array(Ae) -> array_lists(Ae, AL) ; AL=Ae),
    (array(Be) -> array_lists(Be, BL) ; BL=Be),
    clpfd:transpose(AL, BL).




				% NON HYBRID CLPFD CONSTRAINTS JUST GENERALIZED TO ARRAYS AND LISTS


%! all_different(+A) 
%
% constraint on arrays similar to library(clpfd) constraint all_different on lists.

all_different(A):-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, List) ; List=Ae),
    clpfd:all_different(List).


%! all_distinct(A)
%
% constraint on arrays or lists using library(clpfd) constraint all_distinct on list.

all_distinct(A):-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, List) ; List=Ae),
    clpfd:all_distinct(List).



%! lex_chain(+A) 
%
% constraint on array or list A similar to library(clpfd) constraint lex_chain on lists. 

lex_chain(A) :-
    expand(A, Ae),
    (array(Ae) -> array_lists(Ae, Lists) ; Lists=Ae),
    clpfd:lex_chain(Lists).


%! tuples_in(+A, +B) 
%
% constraint on arrays or lists A and B similar to library(clpfd) constraint tuples_in on lists. 

tuples_in(A, B) :-
    expand(A, Ae),
    expand(B, Be),
     (array(Ae) -> array_lists(Ae, AL) ; AL=Ae),
     (array(Be) -> array_lists(Be, BL) ; BL=Be),
     clpfd:tuples_in(AL, BL).

%! serialized(+A, +B) 
%
% constraint on arrays or lists A and B similar to library(clpfd) constraint serialized on lists. 

serialized(A, B) :-
    expand(A, Ae),
    expand(B, Be),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    (array(Be) -> array_list(Be, BL) ; BL=Be),
    clpfd:serialized(AL, BL).

%! global_cardinality(+A, +B) 
%
% constraint on array or list A similar to library(clpfd) constraint global_cardinality on lists. 

global_cardinality(A, B) :-
    expand(A, Ae),
    expand(B, Be),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:global_cardinality(AL, Be).

%! global_cardinality(+A, +B, +C) 
%
% constraint on array or list A similar to library(clpfd) constraint global_cardinality on lists. 

global_cardinality(A, B, C) :-
    expand(A, Ae),
    expand(B, Be),
    expand(C, Ce),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:global_cardinality(AL, Be, Ce).

%! circuit(+A) 
%
% constraint on array or list A similar to library(clpfd) constraint circuit on lists. 

circuit(A) :-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:circuit(AL).

%! cumulative(+A) 
%
% constraint on array or list A similar to library(clpfd) constraint cumulative on lists. 

cumulative(A) :-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:cumulative(AL).

%! cumulative(+A, +O) 
%
% constraint on array or list A with options O similar to library(clpfd) constraint cumulative on lists. 

cumulative(A, B) :-
    expand(A, Ae),
    expand(B, Be),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:cumulative(AL, Be).

%! disjoint2(+A) 
%
% constraint on array or list A similar to library(clpfd) constraint disjoint2 on lists. 

disjoint2(A) :-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:disjoint2(AL).

%! element(+I, +A, +V) 
%
% constraint on array or list B similar to library(clpfd) constraint element on list. 

element(A, B, C) :-
    expand(A, Ae),
    expand(B, Be),
    expand(C, Ce),
    (array(Be) -> array_list(Be, BL) ; BL=Be),
    clpfd:element(Ae, BL, Ce).

%! automaton(+A, +B, +C) 
%
% constraint on arrays or lists A, B, C similar to library(clpfd) constraint automaton on lists. 

automaton(A, B, C) :-
    expand(A, Ae),
    expand(B, Be),
    expand(C, Ce),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    (array(Be) -> array_list(Be, BL) ; BL=Be),
    (array(Ce) -> array_list(Ce, CL) ; CL=Ce),
    clpfd:automaton(AL, BL, CL).


%! chain(+A, +B) 
%
% constraint on array or list A similar to library(clpfd) constraint chain on lists. 

chain(A, B) :-
    expand(A, Ae),
    expand(B, Be),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:chain(AL, Be).



				% CLPFD ENUMERATION


%! label(Var) 
%
% value enumeration predicate on integer arrays or lists similar to library(clpfd) label/1 predicate on lists. 

label(A) :-
    expand(A, Ae),
    (array(Ae) -> array_list(Ae, AL) ; AL=Ae),
    clpfd:label(AL).


%! labeling(Options, Vars) 
%
% enumeration predicate on integer array or list Vars, similar to library(clpfd) labeling/2 predicate on lists,
% but with a new option trace or trace(VarNames) to create the search tree of variable instantiations perfomed by labeling,
% to visualize it with search_tree_text/0, search_tree_term/1, search_tree_latex/1, search_tree_tikz/1 of library(tracesearch).
%
% VarNames can be a list of same length as Vars, or an atom like 'Q' in which case the variables will be named Q1,...,QN.
% ==
% ?- queens(4, Queens), search_tree_text.
% labeling([Q1,Q2,Q3,Q4])
%  Q1=1
%   Q2=3
%   Q2$\neq$3
%  Q1$\neq$1
%   Q1=2
%    [2,4,1,3]
% Queens = array(2, 4, 1, 3) ;
% labeling([Q1,Q2,Q3,Q4])
%  Q1=1
%   Q2=3
%   Q2$\neq$3
%  Q1$\neq$1
%   Q1=2
%    [2,4,1,3]
%   Q1$\neq$2
%    Q1=3
%     [3,1,4,2]
% Queens = array(3, 1, 4, 2) ;
% false.
% ==

labeling(Options, A) :-
    expand(A, Ae),
    expand(Options, Oe), % just in case !
    (array(Ae) -> array_list(Ae, Vars) ; Vars=Ae),
    option_trace(Oe, O, Vars),
    clpfd:labeling(O, Vars),
    add_node(Vars). % success node


option_trace(Options, Opt, Vars) :-
    (member(trace(Arg), Options)
    ->
     delete(Options, trace(_), Opt),
     length(Vars, N),
     (atom(Arg)
     ->
      findall(Name, (between(1, N, I), with_output_to(atom(Name), (write(Arg),write(I)))), VarNames)
     ;
      must_be(list, Arg),
      length(Arg, M),
      (M=N
      ->
       VarNames=Arg
      ;
       throw(error(labeling_trace_option(Arg, Vars))))),
     reverse(VarNames, Rev),
     b_setval(labeling_variable_names_reversed, Rev),
     start_tracing,
     add_node(labeling(VarNames))
    ;
     (member(trace, Options)
     ->
      delete(Options, trace, Opt),
      b_setval(labeling_variable_names_reversed, []),
      start_tracing,
      add_node(labeling(Vars))
     ;
      stop_tracing,
      Opt=Options)).

				% FIRST PATCH TO CLPFD: LABEL/5 in order to interpret new trace(Goal) option 


:- abolish(clpfd:choice_order_variable/7).

clpfd:choice_order_variable(step, Order, Var, Vars, Vars0, Selection, Consistency) :-
         clpfd:fd_get(Var, Dom, _),
         clpfd:order_dom_next(Order, Dom, Next),
	 
	 find_varname(Var, Vars, AtomVar), %FF
         (   add_node(AtomVar = Next), %FF
	     Var = Next,
             clpfd:label(Vars, Selection, Order, step, Consistency)
         ;   
	     with_output_to(atom(Atom), (write(AtomVar),write('$\\neq$'),write(Next))), %FF
	     add_node(Atom),	%FF
	     clpfd:neq_num(Var, Next),
             clpfd:do_queue,
             clpfd:label(Vars0, Selection, Order, step, Consistency)
         ).

clpfd:choice_order_variable(enum, Order, Var, Vars, _Vars0, Selection, Consistency) :-
         clpfd:fd_get(Var, Dom0, _),
	 find_varname(Var, Vars, AtomVar), %FF
	 add_node(enum(AtomVar)), %FF 
         clpfd:domain_direction_element(Dom0, Order, Var),
	 add_node(AtomVar = Var), %FF 
         clpfd:label(Vars, Selection, Order, enum, Consistency).

clpfd:choice_order_variable(bisect, Order, Var, Vars, Vars0, Selection, Consistency) :-
         clpfd:fd_get(Var, Dom, _),
         clpfd:domain_infimum(Dom, n(I)),
         clpfd:domain_supremum(Dom, n(S)),
         Mid0 is (I + S) // 2,
         (   Mid0 =:= S -> Mid is Mid0 - 1 ; Mid = Mid0 ),
 	 find_varname(Var, Vars, AtomVar), %FF
	 concat_atom([AtomVar, ' $\\le$ ', Mid], AtomLE), %FF
	 concat_atom([AtomVar, ' $\\gt$ ', Mid], AtomGT), %FF
         (   Order == up -> ( (add_node(AtomLE), Var #=< Mid) ; (add_node(AtomGT), Var #> Mid) ) %FF
         ;   Order == down -> ( (add_node(AtomGT), Var #> Mid) ; (add_node(AtomLE), Var #=< Mid) ) %FF
         ;   clpfd:domain_error(bisect_up_or_down, Order)
         ),
         clpfd:label(Vars0, Selection, Order, bisect, Consistency).
   
    


% Hack used to find the name of the selected variable to label
% in order to minimize the patch to library(clpfd) hence the multiple calls to reverse
% TODO avoid reverse by adding one argument to clpfd:label 
find_varname(Var, Vars, VarName):-
    b_getval(labeling_variable_names_reversed, RNames), % reversed original variable names
    (RNames=[]
    ->
     term_to_atom(Var, VarName)
    ;
     reverse(Vars, RVars), % reversed tail of non selected original variables
     find_varname(RVars, Var, RNames, VarName)). % search Var in RVars and return its VarName from RNames

find_varname([V | RVars], Var, [N | RNames], VarName) :-
    (Var == V
    ->
     VarName=N
    ;
     find_varname(RVars, Var, RNames, VarName)
    ).
find_varname([], _Var, [VarName | _RNames], VarName). % if not found before





				% SECOND PATCH TO CLPFD: UNIFICATION HOOK in order to prevent duplication of constraint propagators

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Unification hook and constraint projection
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   The reason for this patch is to eliminate the duplication of constraint propagators when unifying variables

Example with library(clpfd):
   
?- L=[A, B], L ins 1..2, A #=< B, bagof(W, member(W, L), L2).
L = L2, L2 = [A, B],
A in 1..2,
B#>=A,
B#>=A,
B#>=A,
B in 1..2.
   
*/

:- abolish(clpfd:attr_unify_hook/2).
clpfd:attr_unify_hook(clpfd_attr(_,_,_,Dom,Ps), Other) :-
%foo(clpfd_attr(_,_,_,Dom,Ps), Other) :-
        (   nonvar(Other) ->
            (   integer(Other) -> true
            ;   type_error(integer, Other)
            ),
            clpfd:domain_contains(Dom, Other),
            clpfd:trigger_props(Ps),
            clpfd:do_queue
        ;
	    % tracing unification of clpfd variables
	    ((b_getval(fdvar_unification_warnings, true),
	      Ps \== fd_props([],[],[]) , get_attr(Other, clpfd, Attr))
	    ->
	     format("Warning: unifying fdvar ~w ~w\n with fdvar domain ~w ~w\n", [Other, Attr, Dom, Ps])
	    ;
	     true),
	    clpfd:fd_get(Other, OD, OPs),
            clpfd:domains_intersection(OD, Dom, Dom1),
            % append_propagators(Ps, OPs, Ps1),  
	    % duplicated constraint propagators after unification, should rather simplify propagators with X=Y
	    union_propagators(Ps, OPs, Ps1),
	    % for the moment just merge propagators without duplicates, 
	    % constraint simplification would need to know the first variable as in clpz.pl not just its attributes
	    clpfd:fd_put(Other, Dom1, Ps1),
            clpfd:trigger_props(Ps1),
            clpfd:do_queue
        ).


union_propagators(fd_props(Gs0,Bs0,Os0), fd_props(Gs1,Bs1,Os1), fd_props(Gs,Bs,Os)) :-
    % sorting the propagators perturbs the queue and can cause severe performance problems
    % maplist(append, [Gs0,Bs0,Os0], [Gs1,Bs1,Os1], [Gs2,Bs2,Os2]),
    % maplist(sort(1, @<), [Gs2, Bs2, Os2], [Gs, Bs, Os]). 
    union_props(Gs0, Gs1, Gs),
    union_props(Bs0, Bs1, Bs),
    union_props(Os0, Os1, Os).


union_props([], P, P).
union_props([P | Ps0], Ps1, Ps):-
    P=propagator(Constraint, Status),
    (contains_prop(Ps1, Constraint, Status)
    ->
     union_props(Ps0, Ps1, Ps)
    ;
     Ps=[P | Pss],
     union_props(Ps0, Ps1, Pss)).

% membership check for constraint propagators
%
% TODO: better than propagator union, add constraint simplification and disentailment check on that event of variable unification 
% e.g. disentailment of all_different all_distinct circuit disjoint2 automaton? chain
% e.g. simplification of sum scalar_product lex_chain serialized global_cardinality cumulative element automaton? transpose? chain

contains_prop([P1 | Ps1], Constraint, Status):-
    arg(1, P1, Constr),
    arg(2, P1, Stat),
    % add disentailement check and constraint simplification here
    (
     Constr==Constraint, Stat=Status % not sure about status unification
    ->
     true
    ;
     contains_prop(Ps1, Constraint, Status)).



				% NEW CLPR CONSTRAINTS


%! float_sum(+VarDomains, +Expr, ?Sum)
% 
% Sum of float Expr defined in comprehension by "in" domain and "where" condition

float_sum(VarDomains, Expr, Sum):-
    aggregation_for(VarDomains, Expr, +, 0.0, '{=}', Sum).


%! float_product(+VarDomains, +Expr, ?Product)
% 
% Product of float Expr defined in comprehension by "in" domain and "where" condition

float_product(VarDomains, Expr, Product):-
    aggregation_for(VarDomains, Expr, *, 1.0, '{=}', Product).



				% PREDICATE NAMES FOR CLPR CONSTRAINTS

clpr_rel(Rel):-
    member(Rel, ['{=}', '{>=}', '{>}', '{=<}', '{<}', '{=\\=}']),
    !.


%! '{=}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 = Expr2}

'{=}'(A, B):-
    {A = B}.

%! '{>=}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 >= Expr2}

'{>=}'(A, B):-
    {A >= B}.

%! '{=<}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 =< Expr2}

'{=<}'(A, B):-
    {A =< B}.

%! '{>}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 > Expr2}

'{>}'(A, B):-
    {A > B}.

%! '{<}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 < Expr2}

'{<}'(A, B):-
    {A < B}.

%! '{=\\=}'(+Expr1, +Expr2)
%
% predicate name for constraint {Expr1 =\= Expr2}

'{=\\=}'(A, B):-
    {A =\= B}.



				% FRONTEND TO CLPR CONSTRAINT ALLOWING SHORTHAND FUNCTIONAL NOTATION

%! {+Constraint}
%
% front-end to library(clpr) {}/1 predicate for allowing shorthand functional notations in Constraint.

{Constraint} :- expand(Constraint, C), clpr:{C}.                      


%! entailed(+Constraint)
%
% front-end to library(clpr) entailed/1 predicate for allowing shorthand functional notations in Constraint.

entailed(Constraint) :- expand(Constraint, C), clpr:entailed(C).

	    
%! maximize(+Expression)
%
% front-end to library(clpr) maximize/1 predicate for allowing shorthand functional notations in Expression.

maximize(Expression) :- expand(Expression, E), clpr:maximize(E).

	    
%! minimize(+Expression)
%
% front-end to library(clpr) minimize/1 predicate for allowing shorthand functional notations in Expression.

minimize(Expression) :- expand(Expression, E), clpr:minimize(E).


%! inf(+Expression, -Inf)
%
% front-end to library(clpr) inf/2 predicate for allowing shorthand functional notations in Expression.

inf(Expression, Inf) :- expand(Expression, E), clpr:inf(E, Inf).


%! sup(+Expression, -Sup)
%
% front-end to library(clpr) sup/2 predicate for allowing shorthand functional notations in Expression.

sup(Expression, Sup) :- expand(Expression, E), clpr:sup(E, Sup).


%! bb_inf(+Ints, +Expression, -Inf, -Vertex, +Eps)
%
% front-end to library(clpr) bb_inf/5 predicate for allowing shorthand functional notations in Expression.

bb_inf(Ints, Expression, Inf, Vertex, Eps) :- expand(Expression, E), clpr:bb_inf(Ints, E, Inf, Vertex, Eps).


%! bb_inf(+Ints, +Expression, -Inf, -Vertex)
%
% front-end to library(clpr) bb_inf/4 predicate for allowing shorthand functional notations in Expression.

%bb_inf(Ints, Expression, Inf, Vertex) :- expand(Expression, E), clpr:bb_inf(Ints, E, Inf, Vertex).


%! bb_inf(+Ints, +Expression, -Inf)
%
% front-end to library(clpr) bb_inf/3 predicate for allowing shorthand functional notations in Expression.

bb_inf(Ints, Expression, Inf) :- expand(Expression, E), clpr:bb_inf(Ints, E, Inf).


% % dump(+Target, +Newvars, -CodedAnswer)
% %
% % call to library(clpr) dump/3 predicate.

% dump(Target, Newvars, CodedAnswer) :- clpr:dump(Target, Newvars, CodedAnswer).

% :- reexport(library(clpfd),
% 	      except([
% 		      (#>)/2,
% 		      (#<)/2,
% 		      (#>=)/2,
% 		      (#=<)/2,
% 		      (#=)/2,
% 		      (#\=)/2,
% 		      (#\)/1,
% 		      (#<==>)/2,
% 		      (#==>)/2,
% 		      (#<==)/2,
% 		      (#\/)/2,
% 		      (#\)/2,
% 		      (#/\)/2,
% 		      in/2,
% 		      ins/2,
% 		      all_different/1,
% 		      all_distinct/1,
% 		      sum/3,
% 		      scalar_product/4,
% 		      tuples_in/2,
% 		      labeling/2,
% 		      label/1,
% 		      lex_chain/1,
% 		      serialized/2,
% 		      global_cardinality/2,
% 		      global_cardinality/3,
% 		      circuit/1,
% 		      cumulative/1,
% 		      cumulative/2,
% 		      disjoint2/1,
% 		      element/3,
% 		      automaton/3,
% 		      transpose/2,
% 		      chain/2,
% 		      op(450, xfx, ..),
% 		      op(500, xfx, \/)
% 		     ]
% 		    )).

% :- op(501, xfx, ..).
% :- op(502, yfx, \/).
   

% :- reexport(library(clpr),
% 	      except([
% 		      {}/1,
% 		      entailed/1,
% 		      maximize/1,
% 		      minimize/1,
% 		      inf/2,
% 		      sup/2, 
% 		      bb_inf/3,
% 		      %bb_inf/4,
% 		      bb_inf/5
% 		     ]
% 		    )).
