/*
  tikz.pl
  
@author Francois Fages
@email Francois.Fages@inria.fr
@license LGPL-2
@version 0.0.1 

  General purpose predicates for drawing the tree structure of a term, in LaTeX tikz (for creating a picture in pdf) or in text.
  
*/


:- module(
	  tikz,
	  [
	   term_to_text/1,
	   term_to_text/2,
	   
	   term_to_tikz/1,
	   term_to_tikz/2,
	   
	   term_to_latex/1,
	   term_to_latex/2
	  ]
	 ).

/** <module>  General purpose predicates for drawing the tree structure of a term, in LaTeX tikz (for creating a picture in pdf) or text.

@author Francois Fages
@version 0.0.1

==
?- term_to_text(this(is, not, a(pipe))).
this
 is
 not
 a
  pipe
true.
==
  
  Used in library(tracesearch) to draw search trees with predicate search_tree_tikz/1
  
  Used in library((clp) and library(modeling) to draw labeling search trees with a new option trace/0  added to predicate labeling/2.
E.g. traced search tree for the 4 queens problem and written in tikz:
==
\begin{tikzpicture}[
	   ->,
	   level/.style={sibling distance=6cm/#1}, % sensitive parameter to adjust manually
	   level distance=2cm
	  ]
\node {{labeling([x1,x2,x3,x4])}}
child {
 node {{x1=1}}
 child {
  node {{x2=3}}
 }
 child {
  node {{x2$\neq$3}}
 }
}
child {
 node {{x1$\neq$1}}
 child {
  node {{x1=2}}
  child {
   node {{[2,4,1,3]}}
  }
 }
 child {
  node {{x1$\neq$2}}
  child {
   node {{x1=3}}
   child {
    node {{[3,1,4,2]}}
   }
  }
  child {
   node {{x1$\neq$3}}
   child {
    node {{x2=1}}
   }
   child {
    node {{x2$\neq$1}}
   }
  }
 }
}
;
\end{tikzpicture}
==

  Note that in the generated LaTeX tikz picture, the distance parameters generally needs to be ajusted manually.
*/

%:- nb_setval(tree_structure_width, 0).

output_stream(Output, Stream) :-
    is_stream(Output)
    ->
    Stream=Output
    ;
    must_be((atom ; string), Output),
    open(Output, write, Stream).


%! term_to_latex(+Term)
%
% same as term_to_latex(current_output, Term).

term_to_latex(Term):-
    term_to_latex(current_output, Term).


%! term_to_latex(+Output, +Term)
%
% writes on Output file or stream a complete LaTeX document that generates an image in pdf of the tree structure of Term.
% Unfortunately tikz does not accept verbatim mode for node labels,
% hence special characters like _ or \ in the writing of Term create LaTeX errors and should be treated first.

term_to_latex(Output, Term):-
    output_stream(Output, Stream),
    format(Stream, "% LaTeX document generated by Prolog library(tikz)\n\\documentclass[tikz,border=10pt]{standalone}\n\\begin{document}\n\\begin{tikzpicture}
  [
   ->,
   level/.style={sibling distance=6cm/#1}, % sensitive parameter to adjust manually
   level distance=2cm
  ]\n\\", []),
    term_to_tikz(Stream, 0, Term),
    format(Stream, ";\n\\end{tikzpicture}\n\\end{document}\n", []),
    close(Stream).

%! term_to_tikz(+Term)
%
% same as term_to_tikz(current_output, Term).

term_to_tikz(Term):-
    term_to_tikz(current_output, Term).


%! term_to_tikz(+Output, +Term)
%
% writes on output Output LaTeX tikz picture code for drawing the tree structure of Term.
% Unfortunately tikz does not accept verbatim mode for node labels,
% hence special characters like _ or \ in the writing of Term create LaTeX errors and should be treated first.

term_to_tikz(Output, Term):-
    output_stream(Output, Stream),
    format(Stream, "% LaTeX tikz code generated by Prolog library(tikz)\n\\begin{tikzpicture}
  [
   ->,
   level/.style={sibling distance=6cm/#1}, % sensitive parameter to adjust manually
   level distance=2cm
  ]\n\\", []),
    term_to_tikz(Stream, 0, Term),
    format(Stream, ";\n\\end{tikzpicture}", []),
    close(Stream).

    
term_to_tikz(Stream, N, Term):-
    foreach(between(1, N, _), write(Stream, ' ')),
    (
     compound(Term)
    ->
     functor(Term, F, A),
     %format(Stream, "node {{\\verb|~w|}}\n", [F]),
     format(Stream, "node {{~w}}\n", [F]),
     N1 is N+1,
     foreach(between(1, A, I), tikz_child(Stream, I, Term, N, N1))
    ;
     %format(Stream, "node {{\\verb|~w|}}\n", [Term])
     format(Stream, "node {{~w}}\n", [Term])
    ).

tikz_child(Stream, I, Term, N, N1) :-
    foreach(between(1, N, _), write(Stream, ' ')), writeln(Stream, "child {"),
    arg(I, Term, Ti),
    term_to_tikz(Stream, N1, Ti),
    foreach(between(1, N, _), write(Stream, ' ')), writeln(Stream, "}").



    
%! term_to_text(+Term)
%
% % draws the tree structure of Term in text.

term_to_text(T):-
    term_to_text(current_output, T).


%! term_to_text(+Output, +Term)
%
% % draws the tree structure of Term in text on Output.

term_to_text(Output, T):-
    output_stream(Output, Stream),
    term_to_text(Stream, 0, T).

term_to_text(Stream, N, Term):-
    foreach(between(1, N, _), write(Stream, ' ')),
    (
     compound(Term)
    ->
     functor(Term, F, A),
     writeln(Stream, F),
     N1 is N+1,
     foreach((between(1, A, I), arg(I, Term, Ti)), term_to_text(Stream, N1, Ti))
    ;
     writeln(Stream, Term)
    ).


